/*****************************************************************************************
 *                                                                                       *
 * OpenSpace                                                                             *
 *                                                                                       *
 * Copyright (c) 2014-2016                                                               *
 *                                                                                       *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of this  *
 * software and associated documentation files (the "Software"), to deal in the Software *
 * without restriction, including without limitation the rights to use, copy, modify,    *
 * merge, publish, distribute, sublicense, and/or sell copies of the Software, and to    *
 * permit persons to whom the Software is furnished to do so, subject to the following   *
 * conditions:                                                                           *
 *                                                                                       *
 * The above copyright notice and this permission notice shall be included in all copies *
 * or substantial portions of the Software.                                              *
 *                                                                                       *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,   *
 * INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A         *
 * PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT    *
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF  *
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE  *
 * OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.                                         *
 ****************************************************************************************/

#ifndef __INTERACTIONHANDLER_H__
#define __INTERACTIONHANDLER_H__

#include <openspace/interaction/keyboardcontroller.h>
#include <openspace/interaction/mousecontroller.h>
#include <openspace/network/parallelconnection.h>
#include <openspace/properties/propertyowner.h>
#include <openspace/properties/stringproperty.h>
#include <openspace/util/keys.h>

#include <mutex>

namespace openspace {

class Camera;
class SceneGraphNode;

namespace interaction {

class InteractionHandler : public properties::PropertyOwner {
public:
    InteractionHandler();

    ~InteractionHandler();

    void setKeyboardController(KeyboardController* controller);
    void setMouseController(MouseController* controller);
    void addController(Controller* controller);

    void lockControls();
    void unlockControls();

    void update(double deltaTime);

    void setFocusNode(SceneGraphNode* node);
    const SceneGraphNode* const focusNode() const;
    void setCamera(Camera* camera);
    const Camera* const camera() const;

    void keyboardCallback(Key key, KeyModifier modifier, KeyAction action);
    void mouseButtonCallback(MouseButton button, MouseAction action);
    void mousePositionCallback(double x, double y);
    void mouseScrollWheelCallback(double pos);

    double deltaTime() const;

    void orbitDelta(const glm::quat& rotation);

    void orbit(const float &dx, const float &dy, const float &dz, const float &dist);

    //void distance(const float &d);

    void rotateDelta(const glm::quat& rotation);

    void distanceDelta(const PowerScaledScalar& distance, size_t iterations = 0);

    void lookAt(const glm::quat& rotation);

    void setRotation(const glm::quat& rotation);

    void resetKeyBindings();
    void bindKey(Key key, KeyModifier modifier, std::string lua);

    void setInteractionSensitivity(float sensitivity);
    float interactionSensitivity() const;

    void setInvertRoll(bool invert);
    bool invertRoll() const;

    void setInvertRotation(bool invert);
    bool invertRotation() const;

    void addKeyframe(const network::datamessagestructures::PositionKeyframe &kf);
    void clearKeyframes();

    /**
    * Returns the Lua library that contains all Lua functions available to affect the
    * interaction. The functions contained are
    * - openspace::luascriptfunctions::setOrigin
    * \return The Lua library that contains all Lua functions available to affect the
    * interaction
    */
    static scripting::ScriptEngine::LuaLibrary luaLibrary();

private:
    friend class Controller;

    InteractionHandler(const InteractionHandler&) = delete;
    InteractionHandler& operator=(const InteractionHandler&) = delete;
    InteractionHandler(InteractionHandler&&) = delete;
    InteractionHandler& operator=(InteractionHandler&&) = delete;

    Camera* _camera;
    SceneGraphNode* _focusNode;

    double _deltaTime;
    std::mutex _mutex;

    bool _validKeyLua;

    std::multimap<KeyWithModifier, std::string > _keyLua;

    float _controllerSensitivity;
    bool _invertRoll;
    bool _invertRotation;

    KeyboardController* _keyboardController;
    MouseController* _mouseController;
    std::vector<Controller*> _controllers;

    properties::StringProperty _origin;
    properties::StringProperty _coordinateSystem;
    
    //remote controller
    std::vector<network::datamessagestructures::PositionKeyframe> _keyframes;
    double _currentKeyframeTime;
    std::mutex _keyframeMutex;
};

} // namespace interaction
} // namespace openspace

#endif // __INTERACTIONHANDLER_H__
